/*   Testcard: A monitor test pattern generator
 *   Copyright (C) 2003-2005 Jason Giglio <jgiglio@acm.org>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   ----------------------------------------------------
 *   Changelog:
 *   0.1: Alpha Version of testpattern, 
 *   	displays single pattern only, at 1024X768
 *   0.2:
 *   	Displays first test pattern, plus smpte color bar pattern
 *   0.3:
 *   	Added generalized gradient function, 
 *   	and a function to use it to draw a complete 
 *   	spectrum of fully saturated colors
 *   1.0 Released, small changes, renamed to testcard
 *   1.1 Made changes to support NTSC SMPTE colors, emulates 7 IRE
 *
 *   You can change the res at the top of main() safely.
 */

#include <unistd.h> //for sleep
#include <stdlib.h>
#include <math.h> //for sin cos
#include <sys/time.h> //for benchmarking

#include "SDL_gfxPrimitives.h"
#include "SDL.h"

//RGBA values, big endian, don't use with SDL, only use with SDL_gfx
#define WHITE	0xFFFFFFFF
#define BLACK	0x000000FF
#define GRAY 	0xB4B4B4FF
#define RED	0xFF0000FF
#define GREEN	0x00FF00FF
#define BLUE	0x0000FFFF
#define CYAN	0x00FFFFFF
#define MAGENTA	0xFF00FFFF
#define YELLOW	0xFFFF00FF

//These are SMPTE NTSC Colors, 7 IRE
#define SBLACK 0x101010FF //my test pattern has this as "black", 
	  //to allow for the blacker than black under the red bar
#define SGRAY 0xB4B4B4FF
#define SYELLOW 0xB4B410FF
#define SCYAN 0x10B4B4FF
#define SGREEN 0x10B410FF
#define SMAGENTA 0xB410B4FF
#define SRED 0xB41010FF
#define SBLUE 0x1010B4FF


#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE (!FALSE)
#endif

typedef int bool;

//functions that full a surface with a complete test pattern
int drawpattern(SDL_Surface *screen, int xres, int yres); //draw the default pattern
int drawsmpte(SDL_Surface *screen, int xres, int yres, bool sat);
int drawspectrum(SDL_Surface *screen, SDL_Rect bbox);
int drawwhitegrille(SDL_Surface *surface, int xres, int yres); //luminance grille for adjusting convergance

//lower level stuff
SDL_Surface* makesurface(int xres, int yres);
void waitforkey();
double benchtimer(double starttime);

//test pattern "primitives"
void drawRow(SDL_Surface*, int offsetx, int offsety, int colors[], int numColors, int gap, int vsize, int xres, int yres); //row of color bars
void hlineSquare(SDL_Surface * screen, int x1, int y1, int size, int step); //a square full of alternating on and off pixels
void lineBurst(SDL_Surface* screen, int x, int y, int radius, int count); //lines radiating from a point
void cross(SDL_Surface* screen, int x, int y, int size); //a cross
int hgradient(SDL_Surface *screen, SDL_Rect bbox, char whichColor, unsigned int initialColor, bool asc); //returns final color



int main(int argc, char *argv[])
{
	//FIXME: take a command line to specify res and which test pattern to shot
	//and also maybe have keys to go back and forth instead of straight through to exit
	//Maybe use the gnu long options package if it makes sense
	int xres=1024;
	int yres=768;
	SDL_Surface *surface;
	SDL_Surface *screen;
	SDL_Rect bbox;
	

	//init SDL video
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) 
	{
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	atexit(SDL_Quit);

	screen = SDL_SetVideoMode(xres, yres, 32, SDL_SWSURFACE|SDL_ANYFORMAT|SDL_FULLSCREEN);
	if ( screen == NULL ) 
	{
		fprintf(stderr, "Unable to set %dx%d video: %s\n", xres, yres, SDL_GetError());
	        exit(1);
	}
	//End init section
  

	//LUMINANCE GRILL
//	 surface=makesurface(xres, yres);
//	 drawwhitegrille(surface, xres, yres);


	//SMTPE PATTERN	
	surface=makesurface(xres, yres);
	drawsmpte(surface, xres, yres, FALSE);
	SDL_BlitSurface(surface, NULL, screen, NULL);
	SDL_UpdateRect(screen,0,0,0,0); //show what we have drawn
	waitforkey();	

	//SMTPE PATTERN, FULL SATURATION	
	surface=makesurface(xres, yres);
	drawsmpte(surface, xres, yres, TRUE);
	SDL_BlitSurface(surface, NULL, screen, NULL);
	SDL_UpdateRect(screen,0,0,0,0); //show what we have drawn
	waitforkey();	


	//CLEAR SCREEN
	SDL_FillRect(screen, NULL, 0x000000);
	SDL_FreeSurface(surface);

	//DRAW GIGS CUSTOM PATTERN	
	surface=makesurface(xres, yres);
	drawpattern(surface, xres, yres);	
	SDL_BlitSurface(surface, NULL, screen, NULL);
	SDL_UpdateRect(screen,0,0,0,0); //show what we have drawn
	waitforkey();	
  
	//CLEAR SCREEN
	SDL_FillRect(screen, NULL, 0x000000);
	SDL_FreeSurface(surface);

	//DRAW SPECTRUM GRADIENT	
	surface=makesurface(xres, yres);
	bbox.x=0;
	bbox.y=0;
	bbox.w=xres;
	bbox.h=yres;
	drawspectrum(surface, bbox);
	SDL_BlitSurface(surface, NULL, screen, NULL);
	SDL_UpdateRect(screen,0,0,0,0); //show what we have drawn
	waitforkey();	
	
	
	return(EXIT_SUCCESS);
} //exit main()


int drawwhitegrille(SDL_Surface *surface, int xres, int yres) //luminance grille for adjusting convergance
{
	//FIXME: we need to make hlinesquare into hlinerect to make this work
//    hlineSquare(screen, int x1, int y1, int size, int step)
}

//draw a spectrum gradient in a rectangle
int drawspectrum(SDL_Surface *screen, SDL_Rect bbox)
{	
	unsigned int lastcolor=0xFF0000FF;
	char varycolor='b';
	int i;
	bool ascdec=TRUE;
	//here's the deal
	//a spectrum gradient goes like this:
	//1. full red, increasing blue
	//2. full blue, decreasing red
	//3. full blue, increasing green
	//4. full green, decreasing blue
	//5. full green, increasing red
	//6. full red, decreasing green
	//This is a continuous thing, but customarily red is drawn at both ends it seems
	bbox.w=bbox.w/6;
	for (i=0; i<6; i++)
	{
		//printf("varycolor: %c lastcolor: %X ascdec: %d\n", varycolor, lastcolor, ascdec);
		lastcolor=hgradient(screen, bbox, varycolor, lastcolor, ascdec);
		ascdec=!ascdec;
		bbox.x+=bbox.w;
		switch (varycolor)
		{
			case 'b': varycolor='r'; break;
			case 'r': varycolor='g'; break;
			case 'g': varycolor='b'; break;
		}
	}
}
	
int hgradient(SDL_Surface *screen, SDL_Rect bbox, char whichColor, unsigned int initialColor, bool asc)
{

	float* counter;
	float red;
	float green;
	float blue;
	
	int offset=0;
	int color;
	float increment=256/(float)bbox.w;
	
	double benchtime;
	benchtime=benchtimer(0);
	switch (whichColor)
	{
		case 'r':
			green=(initialColor & 0x00FF0000)>>16;
			blue=(initialColor & 0x0000FF00)>>8;
			counter=&red;
			break;
		case 'g':
			red=(initialColor & 0xFF000000)>>24;
			blue=(initialColor & 0x0000FF00)>>8;
			counter=&green;
			break;
		case 'b':
			red=(initialColor & 0xFF000000)>>24;
			green=(initialColor & 0x00FF0000)>>16;
			counter=&blue;
			break;
	}

	if (asc)
	{
		for(*counter=0; *counter<256; (*counter)+=increment)
		{
			color=(((int)(red)<<24)|(((int)green)<<16)|(((int)blue)<<8)|(0xFF)); 
			vlineColor(screen, bbox.x+offset, bbox.y, (bbox.y+bbox.h), color); //x, y1, y2
			offset++;
		}
	}
	else
	{
		for(*counter=255; *counter>=0; (*counter)-=increment)
		{
			//color=((red<<24)|(green<<16)|(blue<<8)|(0xFF)); 
			color=(((int)(red)<<24)|(((int)green)<<16)|(((int)blue)<<8)|(0xFF)); 
			vlineColor(screen, bbox.x+offset, bbox.y, (bbox.y+bbox.h), color); //x, y1, y2
			offset++;
		}
	}

	printf("hgradient time: %f\n", benchtimer(benchtime));
	return(color);
}


//benchtime: does a time benchmark.  
//If starttime=0, then return the current time, otherwise, return the difference
//between the starttime and the currenttime

//NOTE:
//there's about 20ms or so function overhead vs doing this inline, but relative times are more
//important than absolute
double benchtimer(double starttime)
{
	struct timeval exetime;
	double benchtime;
	if (starttime==0)
	{
		gettimeofday(&exetime, NULL);
		benchtime=(exetime.tv_sec+((double)exetime.tv_usec/1000000));
		return benchtime;
	}
	else
	{
		gettimeofday(&exetime, NULL);
		benchtime=exetime.tv_sec+((double)exetime.tv_usec/1000000)-starttime;
		return benchtime;
	}

}

int drawsmpte(SDL_Surface *screen, int xres, int yres, bool sat)
{
	
	int counter;
	int x1=0;
	int y1=0;
	int x2=xres*0.142; //each bar takes up about 14.28 percent across
	int y2=yres*0.66; //66% tall
	int *topcolor;
	int topcolors[]={SGRAY, SYELLOW, SCYAN, SGREEN, SMAGENTA, SRED, SBLUE};
	int topcolorssat[]={GRAY, YELLOW, CYAN, GREEN, MAGENTA, RED, BLUE};
	
	//draw the topcolors
	if (sat == TRUE)
		topcolor=topcolorssat;
	else
		topcolor=topcolors;
	int midcolors[]={SBLUE, SBLACK, SMAGENTA, SBLACK, SCYAN, SBLACK, SGRAY};
	int bottomcolor[]={0x001D42FF, 0xEBEBEBFF, 0x2C005CFF, SBLACK};

	for (counter=0;counter<7;counter++)
	{
		boxColor(screen, x1, y1, x2, y2, topcolor[counter]);
		x1=x2;
		x2+=xres*0.143;
	}

	//middle colors
	x1=0; //start over from the left
	y1=y2; //right under the last set
	x2=xres*0.143;
	y2+=yres*0.084; // 8.4 percent tall

	for (counter=0;counter<7;counter++)
	{
		boxColor(screen, x1, y1, x2, y2, midcolors[counter]);
		x1=x2;
		x2+=xres*0.143;
	}

	//bottom colors, at least the first four blocks
	x1=0; //start over from the left
	y1=y2; //right under the last set
	x2=0; //we'll set this before drawing in the loop
	y2+=yres*0.245; // percent tall
	for (counter=0;counter<4;counter++)
	{
		x2+=xres*0.179; //this is before the draw because we aren't drawing a full row
		boxColor(screen, x1, y1, x2, y2, bottomcolor[counter]);
		x1=x2;
	}

	//draw the rest manually, because it's too screwed up to do elegantly
	
	//this is the area under the red bar, it has 1/3rd that is "blacker than black" i.e. "blickity black"
	// 1/3rd that is "Black", and 1/3rd right above black
	x2+=xres*0.048; // 4.6% wide
	boxColor(screen, x1, y1, x2, y2, 0x070707FF);
	x1=x2;
	x2+=xres*0.048;
	boxColor(screen, x1, y1, x2, y2, SBLACK);	
	x1=x2;
	x2+=xres*0.048;
	boxColor(screen, x1, y1, x2, y2, 0x181818FF);	

	//last box
	x1=x2;
	x2=xres;
	boxColor(screen, x1, y1, x2, y2, SBLACK);	
	
}

int drawpattern(SDL_Surface *screen, int xres, int yres)
{
	
	//draw 6 rectangles across the top, RGB primaries, and CMY secondaries
	int offsetx=10; //this is the starting offset for the row
	int offsety=yres/5; //20% from the top
	int primaryColors[]={RED, GREEN, BLUE, CYAN, MAGENTA, YELLOW}; //0xRRGGBBAA
	int numColors=6;
	int gap=10;
	int vsize=50; //horizontal size is automatically determined to fill screen
	int greyColors[]={BLACK, WHITE/10, 0x3F3F3FFF, 0x5F5F5FFF, 0x7F7F7FFF, 0x9F9F9FFF, 0xBFBFBFFF, 0xDFDFDFFF, WHITE};
	int dotSquareSize=yres/6;
	int x;
	int crosssize=20;
	
	xres--;
	yres--; //0 is the first number, so we only get e.g. 0-479
	
	//bounding box
	hlineColor(screen, 0, xres, 0, WHITE); //x1, x2, y
	hlineColor(screen, 0, xres, yres, WHITE);
	vlineColor(screen, 0, yres, 0, WHITE); //x, y1, y2
	vlineColor(screen, xres, yres, 0, WHITE); 


	drawRow(screen, offsetx, offsety, primaryColors, numColors, gap, vsize, xres, yres);	

	//now draw 10 boxes, grey scale from 0 to 100% white in 8 increments
	offsety=vsize+offsety+gap;
	numColors=9;
	gap=0;
	
	drawRow(screen, offsetx, offsety, greyColors, numColors, gap, vsize, xres, yres);	

	//now lets draw a bullseye in the center of the screen
	for (x=1; x<yres/7; x+=3) 
	{
		circleColor(screen, xres/2, yres/2, x, WHITE);
	}
	
	//draw a burst of lines	on either side of it
	lineBurst(screen, xres/4, yres/2, yres/7, yres/5);	
	lineBurst(screen, xres-(xres/4), yres/2, yres/7, yres/8);	

	//draw a hlinesquare in each corner
	hlineSquare(screen, 1, 1, dotSquareSize, 1);	//top left
	hlineSquare(screen, xres-dotSquareSize, 1, dotSquareSize, 1); //top right
	hlineSquare(screen, 1, yres-dotSquareSize, dotSquareSize, 1); //bottom left	
	hlineSquare(screen, xres-dotSquareSize, yres-dotSquareSize, dotSquareSize, 1); //bottom right
	//draw a 50% grey next to the hlinesquare for reference
	boxColor(screen, dotSquareSize, 1, (dotSquareSize*2), dotSquareSize, 0x808080FF);	
	
	//draw some more hlinesquares, but more spaced out lines
	//hlineSquare(screen, dotSquareSize, 1, dotSquareSize, 2); //top left
	//hlineSquare(screen, xres-(2*dotSquareSize), 1, dotSquareSize, 2); //top right
	//hlineSquare(screen, dotSquareSize, yres-dotSquareSize, dotSquareSize, 2); //bottom left
	//hlineSquare(screen, xres-(2*dotSquareSize), yres-dotSquareSize, dotSquareSize, 2); //bottom right
			
	//some crosshatches
	cross(screen, crosssize, yres/2, crosssize); //middle left
	cross(screen, xres-crosssize, yres/2, crosssize); //middle right
	cross(screen, xres/2, crosssize, crosssize); //top center
	cross(screen, xres/2, yres-crosssize, crosssize); //bottom center
	
}//end drawpattern

void cross(SDL_Surface* screen, int x, int y, int size)
{
	hlineColor(screen, x-size, x+size, y, WHITE);
	vlineColor(screen, x, y-size, y+size, WHITE);
}

//draw a burst of lines
void lineBurst(SDL_Surface* screen, int x, int y, int radius, int count)
{
	double angle;
	double pi=3.14159;
	int xoffset;
	int yoffset;
	int increment=360/count;
		
	for (angle=0;angle<360;angle+=increment)
	{
		xoffset=radius*(sin(angle*(pi/180)));
		yoffset=radius*(cos(angle*(pi/180)));
		lineColor(screen, x, y, x+xoffset, y+yoffset, WHITE);
	}
}


//a square of alternating pixels
void hlineSquare(SDL_Surface * screen, int x1, int y1, int size, int step)
{
	int x=x1;
	int y=y1;
	int x2=x1+size;	
	int y2=y1+size;	
	while(y<y2)
	{		
		while(x<x2)
		{
			pixelColor(screen, x, y, WHITE);
			x+=step;
			pixelColor(screen, x, y, BLACK);
			x+=step;
		}
		x=x1; //"CR"
		y++; //"LF"
	}
	
}

void drawRow(SDL_Surface* screen, int offsetx, int offsety, int colors[], int numColors, int gap, int vsize, int xres, int yres)
{
	int boxsize;
	int x;
	boxsize=(xres/numColors)-gap; //-gaps between boxes
	for (x=0; x<numColors; x++)
	{
		boxColor(screen, offsetx, offsety, offsetx+boxsize, offsety+vsize, colors[x]);
		offsetx+=(boxsize+gap);
	}
}

void waitforkey()
{
	 /* Poll for events. SDL_PollEvent() returns 0 when there are no  */
        /* more events on the event queue, our while loop will exit when */
        /* that occurs.                                                  */
	bool quit=FALSE;
	SDL_Event event;
	while(!quit)
	{
		while(SDL_PollEvent(&event))
		{
			/* We are only worried about SDL_KEYDOWN events */
			switch( event.type )
			{
			case SDL_KEYDOWN:
				printf( "Key press detected\n" );
				quit = TRUE;
				break;
			case SDL_QUIT: //window was closed
				quit = TRUE;
				break;

			default:
				break;
			}
		} //end while poll event
	} //end while !quit
} //end wait for key

SDL_Surface* makesurface(int xres, int yres)
{
	SDL_Surface *surface;
	Uint32 rmask, gmask, bmask, amask;

	    /* SDL interprets each pixel as a 32-bit number, so our masks must depend
	     *        on the endianness (byte order) of the machine */
	#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	    rmask = 0xff000000;
	    gmask = 0x00ff0000;
	    bmask = 0x0000ff00;
	    amask = 0x000000ff;
	#else
	    rmask = 0x000000ff;
	    gmask = 0x0000ff00;
	    bmask = 0x00ff0000;
	    amask = 0xff000000;
	#endif    
	
	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, xres, yres, 32, rmask, gmask, bmask, amask);
	if(surface == NULL) 
	{
		fprintf(stderr, "CreateRGBSurface failed: %s\n", SDL_GetError());
		exit(EXIT_FAILURE);
	}
	return surface;
}//end makesurface
